from __future__ import annotations

from typing import Any, List, Tuple

import numpy as np
import tcod

from components.base_component import BaseComponent
from actions import BaseAction, MeleeAction, MovementAction, WaitAction


class BaseAI(BaseAction, BaseComponent):
	entity: Any

	def apply(self) -> None:
		raise NotImplementedError()

	def get_path_to(self, dest_x: int, dest_y: int) -> List[Tuple[int, int]]:
		#copy the walkables
		cost = np.array(self.entity.floor_map.tiles["walkable"], dtype=np.int8)

		#higher numbers deter path-finding this way
		for entity in self.entity.floor_map.entities:
			if not entity.walkable and cost[entity.x, entity.y]:
				cost[entity.x, entity.y] += 10

		graph = tcod.path.SimpleGraph(cost=cost, cardinal=2, diagonal=3)
		pathfinder = tcod.path.Pathfinder(graph)

		pathfinder.add_root((self.entity.x, self.entity.y)) #start pos

		#make the path, omitting the start pos
		path: List[List[int]] = pathfinder.path_to((dest_x, dest_y))[1:].tolist()

		#return the path, after mapping it to a list of tuples
		return [(index[0], index[1]) for index in path]

class AttackOnSight(BaseAI):
	def __init__(self, entity: Actor):
		super().__init__(entity)
		self.path: List[Tuple[int, int]] = []

	def apply(self) -> None:
		target = self.entity.floor_map.player
		xdir = target.x - self.entity.x
		ydir = target.y - self.entity.y
		distance = max(abs(xdir), abs(ydir))

		#if the player can see me, and I'm close enough, attack
		if self.entity.floor_map.visible[self.entity.x, self.entity.y]:
			if distance <= 1:
				return MeleeAction(self.entity, xdir, ydir).apply()
		
			self.path = self.get_path_to(target.x, target.y)
	
		if self.path:
			dest_x, dest_y = self.path.pop(0)
			return MovementAction(
				entity = self.entity,
				xdir = dest_x - self.entity.x,
				ydir = dest_y - self.entity.y,
			).apply()
	
		return WaitAction(self.entity).apply()
